/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

#ifndef COMM_H
#define COMM_H

#include <inttypes.h>
#include <avr/io.h>
#include "global.h"

#ifdef GPSLOG
#include "gps_parse.h"
#endif

#ifdef TRIGGERFRAME_INPUT
#include "triggerframe.h"
#endif

enum MegatunePageIdx
{
    MT_PAGE_IDX_CONFIG = 0,     // config
    MT_PAGE_IDX_HWMAP,          // hw map
    MT_PAGE_IDX_BOOST,          // boost structure
    MT_PAGE_IDX_CONFIGEXT,      // config_ext
    MT_PAGE_IDX_TEMPERATURE,    // ve.mat_enrich table
    MT_PAGE_IDX_A340E,          // at structure
    MT_PAGE_IDX_LAUNCHRPM,	// launchrpm table
    MT_PAGE_IDX_CONFIG_SHARED,

/* following tables all have 2 axis in them (usually RPM and KPA).
* Table size must be specified (8x8, not 8x8+8+8),
* get_bins_pointer() decides which pages share bins => ranges are sent automagically.
* Could be changed, so only LAMBDAVIEW is included and everything after ...IGNADV
*/
    PAGE_IDX_FIRST_RPMKPA,
    // labda size
    MT_PAGE_IDX_LAMBDA = PAGE_IDX_FIRST_RPMKPA, // labmda (lambda_size default 8x8)
    // LAMBDAVIEW is same data as LAMBDA, but 1/x style for retarded tuningprograms
    MT_PAGE_IDX_LAMBDAVIEW,         // labmda (lambda_size default 8x8)
    // ve size
    MT_PAGE_IDX_VE,             	// ve (ve_size default 16x14 in my_make)
    MT_PAGE_IDX_CHARGE,
    MT_PAGE_IDX_MAPCORR,            // map_corr (map_corr_size default 8x8)
    MT_PAGE_IDX_MATRET,

// following tables all share rpmkpa with ign_adv table:
// ign size
    MT_PAGE_IDX_IGNADV,             // ign_adv (adv_size default 12x12)
    MT_PAGE_IDX_BOOST_REF,          // boost ref (adv_size default 12x12)
    MT_PAGE_IDX_1JZVVTI,            // vvti_adv (adv_size default 12x12)
    MT_PAGE_IDX_KNOCK,              // knock ref (adv_size default 12x12)
// please UNDERSTAND page_has_rpmkpa before choosing index for a new table (likely must be placed before VE)
// UPDATE this when adding new table:
    MT_PAGE_IDX_LAST = MT_PAGE_IDX_KNOCK
};
uint8_t MEGATUNE_PAGE_SIZE(uint8_t page);
uint8_t* MEGATUNE_PAGE_POINTER(uint8_t page);
#define MT_PAGE_CRC_PACKET_LEN ((3*(MT_PAGE_IDX_LAST+1))+4)

#define alternative_rtdata()	(config.config12&_BV(ALTERNATIVE_RT_DATA))

#define MT_PAGE_IDX_DUMMY (0xFF)
#define MT_PAGE_SIZE_DUMMY (0)

// only pages that have idx greater or equal MT_PAGE_IDX_VE clone the rpm and kpa axis of the VE table
#define page_has_rpmkpa(pageidx) (pageidx >= PAGE_IDX_FIRST_RPMKPA)

#define MT_CODEVER_LEN (26)

struct privch_t {
	uint8_t a;
	uint8_t b;
};

struct bins_pointer_t {
	uint8_t *p_kpa_range, *p_rpm_range;
	uint8_t kpa_size, rpm_size;
};

struct cmd_queue_t {
  uint8_t (*func)(uint8_t i);	// null is special, means send bytes from (SRAM) private[]
  uint8_t len;    // number of bytes to send
  uint8_t count;  // number of bytes already sent
  union {
    uint8_t * private;
    uint8_t (*privatefunc)(uint8_t i);
  };

  union {
  	struct privch_t privch;
   	// direct access of privw (eg. using privch) is endian-dependent!!!
   	// Do NOT access privw using privch.
   	// Only use privch in case when not accessed as privw
   	// http://www.avrfreaks.net/index.php?module=FreaksFiles&func=viewFile&id=271&showinfo=1
   	// use LO() and HI() macros
	uint16_t privw;
  };
#ifdef SIPR
	uint8_t seq;   // sequence number used in sipr
#endif
};

#define LO(x)	(x & 0xFF)
#define HI(x)	((x >> 8) & 0xFF)

#define MT_LAMBDA_OFFSET 73
#define LAMBDA_INT_OFFSET 200


// We could enumerate instead, but ansi-C promotes enumerations to
// integers (16 bit), no need for that!
// shared variable with TRIGGERFRAME_INPUT data state without overlaping , need reinit on mode changing
#define CMD    0
#define OFFSET 1
#define VALUE  2

#define TRIGGERFRAME_UNESCAPE 0xE0 //comm_state == 0x0E triggerframe input process, next character need unescape

struct com_t {
    uint8_t comm_state;
    uint8_t atomic_helper; // store 2nd byte when "sending" 1st byte of 16 bit value. Change to array if another source needs this (now only SD needs it, the comm functions store value in cmdqueue)
    uint16_t commcrc; // outgoing CRC
	uint16_t incoming_CRC;
	uint8_t primtrig_not_closed;// shows if an headless primary trigger log in progress
    uint8_t comm1_relay_allowed;//if not zero relays com1 to com0 (decrements every 100ms)

#ifdef GPSLOG //allocated size is 24byte
    struct gps_t gps;
#endif
#ifdef TRIGGERFRAME_INPUT //allocated size is RINGBUFLEN + 4
    struct triggerframe_t triggerframe;
#endif

};

#ifdef ONE_WIRE
//status states:
#define ONE_WIRE_LISTENING 0    //default state, wait for incoming 1-wire data
#define ONE_WIRE_CRC_FAIL 1     //1-wire data received, crc check failed
#define ONE_WIRE_CRC_OK 2       //1-wire data received, crc check passed
#define ONE_WIRE_FOUND_CORRECT 3//found everything ok: (1-wire crc and data) or (adc limit)
#define ONE_WIRE_FOUND_INCORRECT 4 //crc ok, but code incorrect, need it for testing ibutton and show it
#define ONE_WIRE_ADC_CHECK 5    //waiting for adc measure ready

#define ONE_WIRE_ADC_LIMIT ( 0x40 )  //1.25V
struct one_wire_t {
    uint8_t status;//see the defines above
    uint8_t tx_cnt;//number of next byte to send
    uint8_t rx_cnt;//number of bytes received
    uint8_t rx_crc;//8bit crc for the incoming 1byte family code and 6 byte serial code
    uint8_t msB;//most significant byte of the serial code
    uint8_t ssB;//second significant byte of the serial code

};
extern struct one_wire_t one_wire;
#endif //ONE_WIRE

extern struct com_t com;
extern uint8_t cmd_queue_head, cmd_queue_tail;
#define cmd_queue_empty() (cmd_queue_head == cmd_queue_tail)

// see TABLES_TO_DUMP* before changing this
// CMD_QUEUE_SIZE must be a power of 2 with the cheap define, no such limitation with the smarter comm_buffer_NOT_full() inline function
#define CMD_QUEUE_SIZE 8
#define comm_buffer_NOT_full() (cmd_queue_head != ((cmd_queue_tail + 1) % CMD_QUEUE_SIZE))
/* smarter than the define: it can accept CMD_QUEUE_SIZE which is not power of 2. But +2 bytes flash.
static inline uint8_t comm_buffer_NOT_full( void )
{
	uint8_t h1 = cmd_queue_head;

	if( h1 ) h1--; // nonzero, no overflow
	else h1 = CMD_QUEUE_SIZE-1;
	return h1 - cmd_queue_tail; // if( cmd_queue_tail == h1 ) return 1; else return 0;
}
*/

extern struct cmd_queue_t cmd_queue[CMD_QUEUE_SIZE];

// WBO2 does not use comm_TX_plainbytes any longer...
// KNOCK_DETAILED_LOG neither, comm_TX_plainbytes() uses mmcbuffer
// #define comm_TX_plain_SIZE (10)
// EXTERN uint8_t comm_TX_plainbytes[comm_TX_plain_SIZE] SECTION(".mmcbuffer"); // ideally after the mmc buffer (so this gets corrupted by stack first)

#ifdef STACKPOINTER_DEBUG
EXTERN uint16_t stackpointer_lowest SECTION(".mmcbuffer"); // the lowest (nonzero) value of SPH,SPL
#endif

void initUART(void);
void initUART1(void);

void comm(uint8_t data);

// keep MT_RTVAR_COUNT a multiples of 4
// (optional filler at the end) so megatune compatibility
// is not broken so often
#define MT_RTVAR_COUNT (56)
#define MT_RTVAR_COUNTMORE (71)
uint8_t sendStuffedSimple(uint8_t cnt);
uint8_t send_page_len_crc_stuffed (uint8_t cnt);
uint8_t SendPage_data_stuffed (uint8_t cnt);
uint8_t SendPage_stuffed( uint8_t i );
/* sendpage error codes*/
#define SENDPAGE_ERROR_NO_PAGEONID 1

uint8_t MTsendRTvar_triggerlog(uint8_t i);
uint8_t MTsendRTvar(uint8_t i, uint8_t caller); //i must be monotonous increasing!!, if caller==0 store ATOMIC_HELPER in cmdqueue, otherwise in com.atomic_helper global var
uint8_t MTtestComm(uint8_t i);
uint8_t MTcodeVer(uint8_t i);

void change_page(uint8_t page);
void get_bins_pointer(uint8_t page, struct bins_pointer_t *bins_pointer);
uint8_t send_page(uint8_t i);
void store_page(uint8_t addr, uint8_t value);
uint8_t sendConfigVar(uint8_t i);
uint8_t allSensorRaw(uint8_t i);
uint8_t ConfigDump(uint8_t cnt);
uint8_t TableDump(uint8_t cnt);
uint8_t TriggerlogPrimtrig(uint8_t cnt);
//TriggerlogAnytrig type defines:
#define TRIGGERFRAME_IGNDEACT 8
#define TRIGGERFRAME_PTRIG_AT0 9
uint8_t TriggerlogAnytrig( uint8_t i );
uint8_t pushfunc(uint8_t (*f)(uint8_t i), uint8_t len);
void pushfuncpar2(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint8_t p2);
void pushfuncpar2W(uint8_t (*f)(uint8_t i), uint8_t len, void * par, uint16_t p2);
void pushfuncpar(uint8_t (*f)(uint8_t i), uint8_t len, void *);
void pushfuncTableDump(uint8_t *p, uint8_t entries);
uint8_t pushfuncsipr(uint8_t (*f)(uint8_t i), uint8_t len, uint8_t seq);

uint8_t srl_direct(uint8_t);
uint8_t srl_direct_hex( uint8_t i );
uint8_t srl_generic(uint8_t);
uint8_t srl_generic_atomic16_crc(uint8_t addr);
uint8_t srl_generic_E(uint8_t addr);
uint8_t srl_generic_P(uint8_t addr);
uint8_t srl_generic_P_upper(uint8_t addr);
uint8_t sendSignature(uint8_t i);

void status_com1(void);
#ifdef ONE_WIRE
void read_1wire(void);
void check_1wire_adc(void);
#endif //ONE_WIRE

/* user defined baudrate */
static inline void change_baudrate( uint8_t newbaud ){
	UBRR0H = 0;
	UBRR0L = newbaud;
}

/* baudrate change pending flag. Ignored when 0,
 * otherwise it contains the new baudrate divisor */
extern uint8_t new_baudrate;

extern uint8_t PROGMEM TABLES_TO_DUMP1[];
extern uint8_t PROGMEM TABLES_TO_DUMP2[];

// Use for debugging
void put_char(uint8_t byte);

/// used  in sipr.c
uint8_t comm_getLambdaView(uint8_t val);
uint8_t comm_setLambdaView(uint8_t val);
#endif
